@@ -4,6 +4,7 @@ See the README and class docstrings for more info.
 import getpass
 import json
 import os
+import re
 from collections import defaultdict
 from datetime import datetime
 from decimal import Decimal, ROUND_HALF_UP
@@ -13,11 +14,13 @@ from urllib.error import URLError
 
 import altair as alt
 import click
+import numpy as np
 import pandas as pd
 from appdirs import user_data_dir
 from canvasapi import Canvas
 from canvasapi.exceptions import InvalidAccessToken, Unauthorized
 from luddite import get_version_pypi
+from tqdm import tqdm
 # Using https://github.com/biqqles/dataclassy instead of dataclasses from
 # stdlibto allow for dataclass inheritance when there are default values. Could
 # use a custom init but it gets messy and the advantage of using dataclasses is
@@ -128,6 +131,10 @@ def cli():
 @click.option('--open-chart', default=None, type=bool, help='Whether to open'
               ' the grade distribution chart automatically.'
               ' Default: Ask at the end')
+@click.option('--filter-assignments', default='.*', type=str, help='Regex to filter'
+              ' which assignments are included in the visualization (case-sensitive).'
+              ' Does not affect the final grade calculation.'
+              ' Default: All (.*)')
 @click.option('--override-campus', default=None, help='Override the automatically'
               ' detected course campus in the CSV file with this text. Default: None')
 @click.option('--override-course', default=None, help='Override the automatically'
@@ -139,8 +146,8 @@ def cli():
 @click.option('--override-subject', default=None, help='Override the automatically'
               ' detected course subject in the CSV file with this text. Default: None')
 def prepare_fsc_grades(course_id, filename, api_url, student_status,
-                       drop_students, drop_threshold, drop_na,
-                       open_chart, override_campus, override_course,
+                       drop_students, drop_threshold, drop_na, open_chart,
+                       filter_assignments, override_campus, override_course,
                        override_section, override_session, override_subject):
     """Prepare course grades for FSC submission.
     \b
@@ -164,7 +171,7 @@ def prepare_fsc_grades(course_id, filename, api_url, student_status,
     """
     fsc_grades = FscGrades(
         course_id, filename, api_url, student_status, drop_students,
-        drop_threshold, drop_na, open_chart, override_campus,
+        drop_threshold, drop_na, open_chart, filter_assignments, override_campus,
         override_course, override_section, override_session, override_subject)
     fsc_grades.connect_to_canvas()
     fsc_grades.connect_to_course()
@@ -175,6 +182,7 @@ def prepare_fsc_grades(course_id, filename, api_url, student_status,
         click.echo('Did not find any assigned grades, exiting.')
     else:
         fsc_grades.save_fsc_grades_to_file()
+        fsc_grades.plot_assignment_scores()
         fsc_grades.plot_fsc_grade_distribution()
         fsc_grades.show_manual_grade_entry_note()
     return
@@ -208,18 +216,6 @@ def show_courses(api_url, filter_, start_date_):
     return
 
 
-@cli.command()
-def z_compare_TA_grading():
-    """**Not yet implemented**"""
-    return
-
-
-@cli.command()
-def z_review_student_grades():
-    """**Not yet implemented**"""
-    return
-
-
 @dataclass
 class CanvasConnection():
     """Parent class for initializing attributes shared between child classes."""
@@ -309,6 +305,7 @@ class FscGrades(CanvasConnection):
     drop_threshold: int
     drop_na: bool
     open_chart: bool
+    filter_assignments: str
     override_campus: str
     override_course: str
     override_section: str
@@ -342,6 +339,7 @@ class FscGrades(CanvasConnection):
         )
         canvas_grades = defaultdict(list)
         for enrollment in enrollments:
+            canvas_grades['User ID'].append(enrollment.user['id'])
             canvas_grades['Student Number'].append(enrollment.user['sis_user_id'])
             surname, preferred_name = enrollment.user['sortable_name'].split(', ')
             canvas_grades['Surname'].append(surname)
@@ -395,18 +393,23 @@ class FscGrades(CanvasConnection):
                 'Remember to post all assignments on Canvas'
                 '\nbefore creating the CSV-file to upload to the FSC.'
                 '\nThere are currently unposted Canvas assignments'
-                '\nthat would change the final score of the following students:\n'
             )
-            click.echo(
-                self.canvas_grades
-                    .query(
-                        '@different_unposted_score == True'
-                    )
-                    .to_markdown(
-                        index=False
-                    )
+            students_with_unposted_score = self.canvas_grades.query(
+                '@different_unposted_score == True'
+            )
+            if students_with_unposted_score.shape[0] > 10:
+                click.echo(
+                    'that would change the final score of '
+                    + click.style(f'{students_with_unposted_score.shape[0]} students.', bold=True)
                 )
-            click.echo()
+                click.echo('Showing the first 10 here:\n')
+                click.echo(students_with_unposted_score[:10].to_markdown(index=False))
+            else:
+                click.echo(
+                    'that would change the final score of the following '
+                    + click.style(f'{students_with_unposted_score.shape[0]} students:\n', bold=True)
+                )
+                click.echo(students_with_unposted_score.to_markdown(index=False))
         return
 
     def drop_student_entries(self):
@@ -420,18 +423,26 @@ class FscGrades(CanvasConnection):
 
         # Drop students that have missing info in any field
         # These are also printed so that it is clear to the user what has happened
-        # and they need to be explicit in disabling the beahvior instead of
+        # and they need to be explicit in disabling the behavior instead of
         # accidentally uploading empty fields to FSC.
         if self.drop_na:
-            dropped_students = dropped_students.append(
-                self.canvas_grades[self.canvas_grades.isna().any(axis=1)])
+            dropped_students = pd.concat([
+                dropped_students,
+                self.canvas_grades[self.canvas_grades.isna().any(axis=1)]
+                ],
+                ignore_index=True
+            )
             self.canvas_grades = self.canvas_grades.dropna()
 
         # Drop students explicitly
         if self.drop_students is not None:
-            dropped_students = dropped_students.append(
+            dropped_students = pd.concat([
+                dropped_students,
                 self.canvas_grades.query(
-                    '`Student Number` in @self.drop_students.split()'))
+                    '`Student Number` in @self.drop_students.split()')
+                ],
+                ignore_index=True
+            )
             self.canvas_grades = self.canvas_grades.query(
                 '`Student Number` not in @self.drop_students.split()').copy()
 
@@ -455,11 +466,11 @@ class FscGrades(CanvasConnection):
         # canvas course code (for UBC courses in general), but there is an override
         # option just in case
         self.campus = 'UBC'
-        self.subject, self.course, self.section, self.session = self.course.course_code.split()
+        self.subject, self.course_name, self.section, self.session = self.course.course_code.split()
         if self.override_campus is not None:
             self.campus = self.override_campus
         if self.override_course is not None:
-            self.course = self.override_course
+            self.course_name = self.override_course
         if self.override_section is not None:
             self.section = self.override_section
         if self.override_session is not None:
@@ -474,7 +485,7 @@ class FscGrades(CanvasConnection):
         # Remove the session number which is only present on Canvas but not FSC
         self.session = self.session[:-1]
         self.fsc_grades[additional_fsc_fields] = (
-            self.campus, self.course, self.section, self.session, self.subject, '', '')
+            self.campus, self.course_name, self.section, self.session, self.subject, '', '')
 
         # Round to whole percentage format since FSC requires that
         # Using Decimal to always round up .5 instead of rounding to even,
@@ -503,7 +514,192 @@ class FscGrades(CanvasConnection):
             self.filename + '.csv',
             index=False
         )
-        click.echo(f'Grades saved to {self.filename}.csv.')
+        click.secho(f'Grades saved to {self.filename}.csv.', bold=True, fg='green')
+        return
+
+    def plot_assignment_scores(self):
+        assignment_regex = re.compile(self.filter_assignments)
+        assignments = [
+            a for a in self.course.get_assignments()
+            if a.published
+              and a.points_possible is not None
+              and a.points_possible > 0
+              and assignment_regex.search(a.name)
+        ]
+
+        assert assignments, (
+            'No assignment names matched'
+            f' the provided regular expression "{self.filter_assignments}"'
+        )
+        assignment_scores_dfs = []
+        click.echo("Downloading assignment scores...")
+        assignment_progress_bar = tqdm(assignments)
+        for assignment in assignment_progress_bar:
+            assignment_progress_bar.set_description(assignment.name)
+            submissions = assignment.get_submissions()
+            assignment_scores = defaultdict(list)
+            for submission in submissions:
+                assignment_scores['User ID'].append(submission.user_id)
+                assignment_scores['Grader ID'].append(submission.grader_id)
+                assignment_scores['Score'].append(
+                    100 * submission.score / assignment.points_possible
+                    if submission.score is not None else None
+                )
+                assignment_scores['Assignment'].append(assignment.name)
+            assignment_scores_dfs.append(pd.DataFrame(assignment_scores))
+        # fillna required on pandas >=1.4.0 due to https://github.com/pandas-dev/pandas/issues/46922
+        assignment_score_df = pd.concat(assignment_scores_dfs, ignore_index=True).fillna(np.nan)
+        # Sometime a negative number is returned for the grader,
+        # which does not make sense, maybe from gradescope?
+        assignment_score_df.loc[assignment_score_df['Grader ID'] < 0, 'Grader ID']  = pd.NA
+
+        user_ids_and_names = {
+            user.id: [user.name, user.sis_user_id]
+            for user in self.course.get_users()
+        }
+        user_ids_and_names_df = pd.DataFrame.from_dict(
+            user_ids_and_names, orient='index', columns=['Name', 'Student Number']
+        )
+        assignment_score_df['Grader'] = assignment_score_df['Grader ID'].map(user_ids_and_names_df['Name'])
+        assignment_score_df['Student'] = assignment_score_df['User ID'].map(user_ids_and_names_df['Name'])
+        assignment_score_df['Student Number'] = assignment_score_df['User ID'].map(
+            user_ids_and_names_df['Student Number']
+        )
+
+        # Using `round` instead of `Decimal` here
+        # since the latter can't deal with a df with a single `None`
+        # and because this is just to show on the assignment scores,
+        # so it does not have to be fairly rounded like the final FSC grades.
+        assignment_score_df['Score'] = assignment_score_df['Score'].round(2)
+        # self.canvas has had dropped students removed at this point
+        # so we can use it to drop from the assignment score as well
+        assignment_score_df = assignment_score_df.query(
+            '`User ID` in @self.canvas_grades["User ID"]'
+        ).copy()
+
+        grader_order = assignment_score_df.groupby(
+            'Grader'
+        )['Score'].mean().sort_values().index.tolist()
+        # assignment_order is only needed because VL does not support maintaining
+        # the orignal order for facets https://github.com/vega/vega-lite/issues/6221
+        assignment_order = assignment_score_df['Assignment'].unique().tolist()
+        height = max(80, len(grader_order) * 20)
+
+        assignment_central_tendencies = alt.Chart(
+                assignment_score_df
+            ).transform_aggregate(
+                Mean='mean(Score)',
+                Median='median(Score)',
+            ).transform_fold(
+                fold=['Mean', 'Median'],
+                as_=['Type', 'Percent Grade']
+            ).mark_point(
+            size=65,
+            shape='diamond',
+            fill='white'
+        ).encode(
+            x='Percent Grade:Q',
+            y=alt.value(height-6),  # -6 to get the bottom of the marker on the axis line
+            color=alt.Color(
+                'Type:N',
+                title='',
+                scale=alt.Scale(range=['coral', 'rebeccapurple']),
+                legend=None
+            ),
+            tooltip=['Type:N', alt.Tooltip('Percent Grade:Q', format='.3g')]
+        )
+
+        # Min height=80 for histograms to look nice
+        # 20 is the default step size for categorical scale
+        self.assignment_distributions = alt.hconcat((
+            alt.Chart(
+                assignment_score_df,
+                height=height,
+            ).mark_bar().encode(
+                x=alt.X('Score', bin=alt.Bin(step=5)),
+                y=alt.Y('count()', title='Student Count'),
+            ) + assignment_central_tendencies
+            ).facet(
+                title=alt.TitleParams(
+                    'Assignment Score Distributions',
+                    subtitle=['Hover over the points to see the exact mean and median score.', ''],
+                    anchor='middle',
+                    dx=25
+                ),
+                facet=alt.Facet('Assignment', title='', sort=assignment_order),
+                columns=1
+            ), alt.Chart(
+                assignment_score_df.reset_index(),
+                height=height + 2,
+            ).mark_boxplot(median={'color': 'black'}).encode(
+                x=alt.X('Score', scale=alt.Scale(zero=False)),
+                y=alt.Y('Grader:N', sort=grader_order, title='', axis=alt.Axis(orient='right')),
+                color=alt.Color('Grader:N', sort=grader_order, legend=None)
+            ).facet(
+                title=alt.TitleParams(
+                    'Comparison Between Graders',
+                    subtitle=['Hover over the box for detailed grader info.', ''],
+                    anchor='middle',
+                    dx=-40
+                ),
+                facet=alt.Facet('Assignment', title='', sort=assignment_order),
+                columns=1
+            ).resolve_scale(
+                y='independent',  # Don't use the same y-axis ticks for each faceted boxplot
+            ),
+            spacing=60
+        ).resolve_scale(
+            color='independent'  # Don't use the mean/median color range for the boxplot
+        )
+
+        # Plot assignment scores
+        assignment_score_df['Assignment scores stdev'] = assignment_score_df['User ID'].map(
+            assignment_score_df.groupby('User ID')['Score'].std()
+        )
+        self.hover = alt.selection_single(
+            fields=['User ID'], on='mouseover', nearest=True, empty='none'
+        )
+        base = alt.Chart(
+                assignment_score_df,
+                title=alt.TitleParams(
+                    'Student Assignment Scores',
+                    subtitle=[
+                        'Hover near a point to highlight a line.',
+                        'Hover directly over a point to view student info.',
+                        'Click the three dots button to the right to save this entire page.',
+                    ],
+                    anchor='middle',
+                    dx=25
+                ),
+            ).mark_point(opacity=0).encode(
+            y=alt.Y('Score', scale=alt.Scale(zero=False), title='Assignment Score (%)'),
+            detail='User ID',
+            x=alt.X('Assignment', title='', sort=assignment_order),
+            # Having the tooltip here instead of in the transformed chart
+            # makes it work with nearest,
+            # but significantly slows down the higlighting of the line
+            # tooltip=['Student', 'Student Number', 'Score'],
+        )
+        width = min(1000, max(400, 80 * assignment_score_df['Assignment'].nunique()))
+        self.assignment_scores = (
+            base.add_selection(
+                self.hover
+            ) + base.mark_line(opacity=0.5, interpolate='monotone').encode(
+                color=alt.Color(
+                    'Assignment scores stdev',
+                    scale=alt.Scale(scheme='cividis', reverse=True),
+                    title=['Stdev between', 'Assignments']
+                )
+            ) + (base.mark_line(size=3, interpolate='monotone', color='maroon')
+            + base.mark_circle(size=70, color='maroon', opacity=1).encode(
+                tooltip=['Student', 'Student Number', 'Score'],
+                )
+            ).transform_filter(
+                self.hover
+            )).properties(
+                width=width,
+                height=280,
+            ).interactive()
         return
 
     def plot_fsc_grade_distribution(self):
@@ -519,7 +715,7 @@ class FscGrades(CanvasConnection):
                 }
             # Combine the rounded and raw *unposted* scores 
             ).melt(
-                id_vars=['Preferred Name', 'Surname', 'Student Number'],
+                id_vars=['Preferred Name', 'Surname', 'Student Number', 'User ID'],
                 value_vars=['FSC Rounded', 'Exact'],
                 value_name='Unposted Grade',
                 var_name='Percent Type'
@@ -532,14 +728,14 @@ class FscGrades(CanvasConnection):
                 }
             # Combine the rounded and raw *posted* scores 
             ).melt(
-                id_vars=['Preferred Name', 'Surname', 'Student Number'],
+                id_vars=['Preferred Name', 'Surname', 'Student Number', 'User ID'],
                 value_vars=['FSC Rounded', 'Exact'],
                 value_name='Posted Grade',
                 var_name='Percent Type'
             )
         # Combine the posted and unposted scores
         ).melt(
-            id_vars=['Preferred Name', 'Surname', 'Student Number', 'Percent Type'],
+            id_vars=['Preferred Name', 'Surname', 'Student Number', 'User ID', 'Percent Type'],
             value_vars=['Unposted Grade', 'Posted Grade'],
             value_name='Percent Grade',
             var_name='Grade Status'
@@ -553,7 +749,7 @@ class FscGrades(CanvasConnection):
         grade_status_selection = alt.selection_single(
             fields=['Grade Status'],
             bind=grade_status_dropdown,
-            init={'Grade Status': 'Posted Grade'}
+            init={'Grade Status': 'Unposted Grade'}
         )
         percent_type_dropdown = alt.binding_select(
             options=['FSC Rounded', 'Exact'],
@@ -562,13 +758,13 @@ class FscGrades(CanvasConnection):
         percent_type_selection = alt.selection_single(
             fields=['Percent Type'],
             bind=percent_type_dropdown,
-            init={'Percent Type': 'FSC Rounded'}
+            init={'Percent Type': 'Exact'}
         )
 
         # Plot distribution
         hist = alt.Chart(self.fsc_grades_for_viz, height=200).mark_bar().encode(
-            alt.X('Percent Grade', bin=alt.Bin(maxbins=15), title='', axis=alt.Axis(labels=False)),
-            alt.Y('count()', title='Number of students')
+            alt.X('Percent Grade', bin=alt.Bin(step=5), title='', axis=alt.Axis(labels=False)),
+            alt.Y('count()', title='Student Count')
         )
 
         # Plot all observations
@@ -580,7 +776,7 @@ class FscGrades(CanvasConnection):
             Name='datum["Preferred Name"] + " " + datum["Surname"]'
         ).encode(
             alt.X('Percent Grade',
-                title='Percent Grade',
+                title='Final Percent Grade',
                 scale=alt.Scale(
                     zero=False,
                     nice=False,
@@ -597,8 +793,14 @@ class FscGrades(CanvasConnection):
                     grid=False
                 )
             ),
-            alt.Tooltip(['Name:N', 'Student Number', 'Percent Grade'])
-        ).interactive()
+            alt.Tooltip(['Name:N', 'Student Number', 'Percent Grade']),
+        )
+
+        strip_overlay = strip.mark_circle(size=80, opacity=1).encode(
+            color=alt.value('maroon')
+        ).transform_filter(
+            self.hover
+        )
 
         # Plot central tendencies
         central_tendencies = alt.Chart(
@@ -621,37 +823,41 @@ class FscGrades(CanvasConnection):
                 'Type:N',
                 title='',
                 scale=alt.Scale(range=['coral', 'rebeccapurple']),
-                legend=alt.Legend(legendY=225, legendX=415, orient='none')
+                legend=alt.Legend(legendY=300, legendX=295, orient='none', columns=2)
             ),
             tooltip=['Type:N', alt.Tooltip('Percent Grade:Q', format='.3g')]
         )
 
         # Add instructions
         title = alt.TitleParams(
-            text=f'Grade distribution {self.subject} {self.course}',
+            text=f'Final Grade Distribution {self.subject} {self.course_name}',
             subtitle=[
-                'Hover over the points to see the student name and grade.',
+                'Hover near a point to highlight it and view student info.',
                 'Zoom with the mouse wheel and double click to reset the view.',
-                'Click the three dots button to the right to save the plot.']
+                'Changes in the dropdown menus below only affect this chart'
+            ],
+            anchor='middle',
+            dx=25
         )
 
         # Concatenate, add filters, and save the chart
         chart_filename = self.filename + '.html'
-        alt.vconcat(
+        alt.vconcat((alt.vconcat(
             hist,
             # strip on top so that individual observations are always visible
-            central_tendencies + strip,
+            strip.add_selection(self.hover).interactive() + strip_overlay + central_tendencies,
             spacing=0
         ).properties(
             title=title
         ).resolve_scale(
             x='shared'
-        ).configure_view(
-            strokeWidth=0
         ).transform_filter(
                 percent_type_selection & grade_status_selection
         ).add_selection(
             percent_type_selection, grade_status_selection
+        ) | self.assignment_scores), self.assignment_distributions, spacing=60
+        ).resolve_scale(
+            color='independent'
         ).save(
             chart_filename
         )
@@ -661,10 +867,14 @@ class FscGrades(CanvasConnection):
                 '\n<style>'
                 '\n    form.vega-bindings {'
                 '\n        font-family: sans-serif;'
+                '\n        font-size: 12px;'
+                '\n        position: absolute;'
+                '\n        left: 60px;'
+                '\n        top: 395px;'
                 '\n    }'
                 '\n</style>'
             )
-        click.echo(f'Grade distribution chart saved to {chart_filename}.')
+        click.secho(f'Grade distribution chart saved to {chart_filename}.', bold=True, fg='green')
         if self.open_chart or self.open_chart is None and click.confirm(
                 'Open grade distribution chart?', default=True):
             click.launch(chart_filename)
@@ -674,12 +884,13 @@ class FscGrades(CanvasConnection):
         """Show disclaimer and note about manual grade entry of student standings."""
         click.secho('\nNOTE', fg='yellow', bold=True)
         click.echo(
-            '1. Make sure to check the output above for any warning and notes.'
+            '1. Check the output above for any warnings and notes.'
             '\n2. The saved CSV file should automatically be correctly formatted,'
-            '\n   but it is ' + click.style('your responsibility', bold=True) + ' to double check in case there are unexpected changes'
+            '\n   but it is ' + click.style('your responsibility to double check', bold=True) +
+            '\n   since there could be unexpected changes'
             '\n   to how UBC inputs course info on Canvas.'
             '\n3. If you have students that did not take the final exam'
-            '\n   or with a thesis in progress, you will need to enter this info manually.'
+            '\n   you will need to enter this info manually.'
             '\n   Please see https://facultystaff.students.ubc.ca/files/FSCUserGuide.pdf'
             '\n   under "Acceptable Values for Grades Entry" for how to modify the CSV file.')
         return
